import requests
from abc import abstractmethod
from typing import List

from .Objects import Color

BASE_URL = "https://developer-api.govee.com"

class GoveeDevice:
    """
    Class representing a Govee device

    :param str device: the unique device id.
    :param str model: the model of the device.
    """
    _model: str
    _device: str
    _name: str

    _controllable: bool
    _retrievable: bool

    _key: str

    _state: bool = None

    def __init__(self,
                 model: str = "",
                 device: str = "",
                 name: str = "",
                 controllable: bool = False,
                 retrievable: bool = False,
                 key: str = ""):
        self._model = model
        self._device = device
        self._name = name

        self._controllable = controllable
        self._retrievable = retrievable

        self._key = key

        self._update()

    def _make_request(self, method: str, url: str, *args, **kwargs) -> requests.Response:
        if not kwargs.get("headers", None):
            kwargs["headers"] = {
                "Govee-API-Key": self._key
            }
        response = requests.request(method, url, *args, **kwargs, timeout=60)
        return response.json()

    @abstractmethod
    def _command(self, command):
        pass

    @abstractmethod
    def _update(self):
        pass

    @property
    def model(self):
        """Device Model"""
        return self._model

    @property
    def device(self):
        """Device ID"""
        return self._device

    @property
    def name(self):
        """The Devices Name"""
        return self._name

    @property
    def controllable(self):
        """Whether the device is controllable"""
        return self._controllable

    @property
    def retrievable(self):
        """Whether the device is retrievable"""
        return self._retrievable

    @property
    def on(self):
        """Whether the light is on"""
        return self._state

    @property
    def off(self):
        """Whether the light is off"""
        return not self._state

    def turn_on(self):
        """Turn the light on"""
        self._command(
            {
                "name": "turn",
                "value": "on"
            }
        )
        self._state = True

    def turn_off(self):
        """Turn the light off"""
        self._command(
            {
                "name": "turn",
                "value": "off"
            }
        )
        self._state = False

    def __str__(self):
        return f"<{self.name} [device = {self.device}, model = {self.model}]>"

    def __repr__(self):
        return f"<{self.name} [device = {self.device}, model = {self.model}]>"



class GoveeLight(GoveeDevice):
    """
    Class representing a Govee light.

    You should not need to create this class, I recommend using :meth:`Govee.get_devices` to get a list of devices.
    
    :param str device: the unique device id.
    :param str model: the model of the device.
    :param str name: the name of the device.
    :param bool controllable: whether the device is controllable.
    :param bool retrievable: whether the device is retrievable.
    :param str key: your API key for Govee.
    """

    _online: bool = True
    _brightness: int = 0
    _color: Color = Color(0,0,0)

    def _command(self, command):
        self._make_request(
            "PUT",
            f"{BASE_URL}/v1/devices/control",
            json = {
                "device": self._device,
                "model": self._model,
                "cmd": command
            }
        )

    def _update(self):
        response = self._make_request(
            "GET",
            f"{BASE_URL}/v1/devices/state",
            params={
                "device": self._device,
                "model": self._model
            }
        )
        data = {}
        for i in response["data"]["properties"]:
            for k, v in i.items():
                data[k] = v
        self._state = data.get("powerState") == 'on'
        self._brightness = data.get("brightness", 0)
        self._color = Color(**data.get("color", {"r":0,"g":0,"b":0}))

    def set_brightness(self, brightness: int):
        """Set the brightness of the light (1-100)"""
        self._command({
            "name": "brightness",
            "value": brightness % 101
        })
        self._brightness = brightness % 101

    def set_color(self, r: int, g: int, b: int):
        """Set the color of the light"""
        self._command({
            "name": "color",
            "value": {
                "r": r % 256,
                "g": g % 256,
                "b": b % 256
            }
        })
        self._color = Color(r%256,g%256,b%256)

class GoveeAppliance(GoveeDevice):
    """
    Class representing a Govee appliance.
    
    Note:
        There is currently no good API implementation for Govee appliances.
    """

    def _command(self, command):
        """
        Sends a command to control the Govee appliance.

        Args:
            command (str): The command to send to the Govee appliance.

        Returns:
            None
        """
        self._make_request(
            "PUT",
            f"{BASE_URL}/v1/appliance/devices/control",
            json={
                "device": self._device,
                "model": self._model,
                "cmd": command
            }
        )

    def _update(self):
        """
        Updates the Govee appliance.

        Note:
            There is currently no API implementation for updating the Govee appliance.

        Returns:
            None
        """
        pass


class Govee:
    """
    Represents an instance connected to the Govee API.

    :param str key: Your API Key for Govee.
    """
    _key: str
    _devices: List[GoveeDevice]
    _device_rate_limits: List

    def __init__(self, key: str):
        self._key = key
        self._devices = []

    def _make_request(self, method: str, url: str, *args, **kwargs):
        if not kwargs.get("headers", None):
            kwargs["headers"] = {
                "Govee-API-Key": self._key
            }
        response = requests.request(method, url, *args, **kwargs, timeout=60)
        if response.status_code == 200:
            return response.json()

    def get_devices(self, *, update: bool = False):
        """
        Retrieves the list of devices associated with the Govee account.

        :param bool update: If True, force an update of the device list.
        :return: List of GoveeDevice objects.
        :rtype: List[GoveeDevice]
        """
        if len(self._devices) == 0 or update:
            data = self._make_request(
                "GET",
                f"{BASE_URL}/v1/devices"
            )
            for device in data["data"]["devices"]:
                self._devices.append(
                    GoveeLight(
                        device["model"],
                        device["device"],
                        device["deviceName"],
                        device["controllable"],
                        device["retrievable"],
                        self._key
                    )
                )
        return self._devices
        
    def get_device_by_name(self, name: str) -> GoveeDevice | None:
        """
        Retrieves a device by its name.

        :param str name: The name of the device.
        :return: The GoveeDevice object if found, otherwise None.
        :rtype: GoveeDevice | None
        """
        for device in self._devices:
            if device.name == name:
                return device
        return None
    
    def get_device_by_model(self, model: str) -> GoveeDevice | None:
        """
        Retrieves a device by its model.

        :param str model: The model of the device.
        :return: The GoveeDevice object if found, otherwise None.
        :rtype: GoveeDevice | None
        """
        for device in self._devices:
            if device.model == model:
                return device
        return None
    
    def get_device_by_address(self, address: str) -> GoveeDevice | None:
        """
        Retrieves a device by its address.

        :param str address: The address of the device.
        :return: The GoveeDevice object if found, otherwise None.
        :rtype: GoveeDevice | None
        """
        for device in self._devices:
            if device.device == address:
                return device
        return None


import logging
import socket
import json
import time

from typing import List, Callable
from .Objects import Color

_LOG = logging.getLogger(__name__)

_MULTICAST_GROUP = '239.255.255.250'
_MULTICAST_PORT = 4001
_RESPONSE_IP = '0.0.0.0'
_RESPONSE_PORT = 4002
_DEVICE_PORT = 4003

class _Listener:

    def _listen_for_response(self, address, port, *,
                            timeout: int = None,
                            callback: Callable[[dict, str], None] = None
                            ) -> dict | None:
        _LOG.info("Listening for response")
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setblocking(0)
        sock.bind((address, port))
        if timeout:
            start = time.time()
        while True:
            try:
                recieved, resp_addr = sock.recvfrom(1024)
                recieved.decode("utf-8")
                try:
                    data = json.loads(recieved)
                    _LOG.info("Response from %s, %s",resp_addr, data)
                    if timeout:
                        callback(data, resp_addr)
                    else:
                        return data
                except json.JSONDecodeError as e:
                    _LOG.error("Error decoding JSON: %s", e)
            except socket.error:
                pass

            if timeout:
                if time.time() - start > timeout:
                    _LOG.info("Timeout reached. Stopping UDP server.")
                    break

class GoveeDeviceLocal(_Listener):
    """
    Class representing a Govee device

    :param str ip: the IP address of the device.
    :param str device: the unique device id.
    :param str model: the model of the device.
    """
    _ip: str
    _device: str
    _model: str

    _state: bool
    _brightness: int
    _color: Color
    _color_tem: int

    def __init__(self, ip: str, device: str, model: str) -> None:
        self._ip = ip
        self._device = device
        self._model = model

    def _send_request(self, command: dict, *, listen: bool = False) -> dict | None:
        """
        Sends a request to the device.

        :param command: The command to send.
        :param listen: Whether or not to listen for a response.
        """
        data: dict | None = None

        r_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        message = json.dumps({
            "msg": command
        }).encode("utf-8")

        try:
            r_sock.sendto(message, (self._ip, _DEVICE_PORT))
            _LOG.info("Request sent to %s:%s, %s", self._ip, _DEVICE_PORT, command)
        except socket.error:
            _LOG.error("Request Failed")
        finally:
            r_sock.close()
            if listen:
                data = self._listen_for_response(_RESPONSE_IP, _RESPONSE_PORT)
        return data

    def turn_on(self):
        """
        Turns the device on. 

        :return: None
        """

        self._send_request(
            {
                "cmd" : "turn",
                "data": {
                    "value" : 1
                }
            }
        )
        self._state = True

    def turn_off(self):
        """
        Turns the device off.    

        :return: None
        """

        self._send_request(
            {
                "cmd" : "turn",
                "data": {
                    "value" : 0
                }
            }
        )
        self._state = False

    def set_brightness(self, brightness: int):
        """
        Set the brightness of the device.

        :param brightness: The brightness to set (1-100)
        :return: None
        """

        self._send_request({
            "cmd" : "brightness",
            "data": {
                "value": brightness % 101
            }
        })
        self._brightness = brightness % 101

    def set_color(self, r: int, g: int, b: int, temp: int = 0):
        """
        Set the color of the device.
        
        :param r: The red value (0-255)
        :param g: The green value (0-255)
        :param b: The blue value (0-255)
        :param temp: The color temperature (0-100)
        :return: None
        """

        self._send_request({
            "cmd": "colorwc",
            "data": {
                "color" : {
                    "r": r % 256,
                    "g": g % 256,
                    "b": b % 256
                },
                "colorTemInKelvin": temp
            }
        })
        self._color = Color(r%256, g%256, b%256)
        self._color_tem = temp

    def _update_device_state(self):
        """
        Updates the device state.
        """
        response = self._send_request({
            "cmd": "devStatus",
            "data": {}
        }, listen=True)["msg"]["data"]

        self._state = bool(response.get("onOff", 0))
        self._brightness = response.get("brightness", 0)
        self._color = Color(**response["color"]) if "color" in response else Color(0,0,0)
        self._color_tem = response.get("ColorTemInKelvin", 0)

    def update(self):
        """
        Updates the device state.
        """
        self._update_device_state()

    @property
    def on(self):
        """
        Returns whether or not the device is on.
        
        :return: bool
        """
        return self._state

    @property
    def off(self):
        """
        Returns whether or not the device is off.
        
        :return: bool
        """
        return not self._state

    @property
    def ip(self):
        """
        Returns the IP address of the device.
        
        :return: str
        """
        return self._ip

    @property
    def device(self):
        """
        Returns the unique device id.
        
        :return: str
        """
        return self._device

    @property
    def model(self):
        """
        Returns the model of the device.
        
        :return: str
        """
        return self._model

    def __str__(self):
        return f"<LocalGoveeDevice || {self.ip = }|{self.device = }|{self.model = }>"

    def __repr__(self):
        return f"<LocalGoveeDevice || {self.ip = }|{self.device = }|{self.model = }>"

class GoveeLocal(_Listener):
    """
    Class representing a Govee device
    
    :param int timeout: The timeout for the scan request.
    
    :var devices: A list of devices found on the network.
    
    :return: None
    """

    def __init__(self, *, timeout: int = 1):
        self._timeout = timeout
        self.__devices = []

    def get_devices(self):
        """
        Sends a scan request to the network and returns a list of devices found.
        
        :return: List[GoveeDeviceLocal]
        """
        return self._send_scan_request()

    def get_device_by_device(self, device: str) -> GoveeDeviceLocal | None:
        """
        Returns a device by its unique device id.

        :param device: The unique device id.
        
        :return: GoveeDeviceLocal | None
        """

        dev = list(filter(lambda x: x.device == device, self.__devices))
        return dev[0] if len(dev) > 0 else None
    
    def get_device_by_ip(self, ip: str) -> GoveeDeviceLocal | None:
        """
        Returns a device by its IP address.

        :param ip: The IP address of the device.

        :return: GoveeDeviceLocal | None
        """
        dev = list(filter(lambda x: x.ip == ip, self.__devices))
        return dev[0] if len(dev) > 0 else None
    
    def __handle_response(self, response_data, sender_address):
        """
        Handles a response from a device.
        """
        self.__devices.append(GoveeDeviceLocal(
            response_data["msg"]["data"]["ip"],
            response_data["msg"]["data"]["device"],
            response_data["msg"]["data"]["sku"]
        ))
        _LOG.info("Received response from %s: %s", sender_address, response_data)

    def _send_scan_request(self):
        """
        Sends a scan request to the network.
        """
        send_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        send_socket.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
        request_message = json.dumps({
            "msg": {
                "cmd": "scan",
                "data": {
                    "account_topic": "reserve"
                }
            }
        }).encode('utf-8')
        try:
            send_socket.sendto(request_message, (_MULTICAST_GROUP, _MULTICAST_PORT))
            _LOG.info("Request sent to %s:%s", _MULTICAST_GROUP, _MULTICAST_PORT)
        except socket.error as e:
            _LOG.error("Error Sending Request to %s:%s | %s", _MULTICAST_GROUP, _MULTICAST_PORT, e)
        finally:
            send_socket.close()
        self._listen_for_response(_RESPONSE_IP, _RESPONSE_PORT, timeout=1, callback=self.__handle_response)
        for device in self.__devices:
            device.update()
        return self.__devices

    @property
    def devices(self):
        """
        Returns a list of cached devices found on the network.
        
        :return: List[GoveeDeviceLocal]
        """
        return self.__devices


from setuptools import setup

setup(
  name='GoveePy',
  version='0.3',
  description='A Govee API Wrapper for Python',
  url='https://github.com/WhyDoWeLiveWithoutMeaning/GoveePy',
  author="Meaning",
  license="MIT",
  packages=['govee'],
  install_requires=[
    "requests>=2.31.0"
  ],
  classifiers=[
    'Development Status :: 1 - Planning',
    'License :: OSI Approved :: MIT License',  
    'Programming Language :: Python :: 3.12',
    'Typing :: Typed'
  ]
)